#include <stdint.h>
#include "nrf_sdh.h"
#include "nrf_sdh_ant.h"
#include "nrfx_rtc.h"
#include "app_error.h"
#include "ant_interface.h"
#include "ant_parameters.h"
#include "ant_channel_config.h"
#include "ant_search_config.h"

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

// ANT channel config.
#define CHAN_ID_DEV_TYPE               1                            ///< Channel ID, Device Type. 
#define CHAN_ID_DEV_NUM                1                            ///< Channel ID, Device Number. 
#define CHAN_ID_TRANS_TYPE             1                            ///< Channel ID, Transmission type.
#define CHAN_PERIOD                    9687                         ///< Channel Period (in 32 kHz counts).
#define RF_FREQ                        66                           ///< RF Frequency: 2466 MHz. 
#define ANT_CHANNEL_DEFAULT_NETWORK    0x00                         ///< ANT Channel Network.
#define ANT_CHANNEL_NUMBER             0x00                         ///< ANT Channel Number.
#define EXT_TYPE                       0                            ///< Extended assign.
#define APP_ANT_OBSERVER_PRIO          1                            ///< Application's ANT observer priority. You shouldn't need to modify this value.

// Timesync Page Numbers
#define TIME_SYNC_PAGE                 0x01 
#define INVALIDATION_BYTE              0xFF   

// RTC1 instance
const nrfx_rtc_t m_rtc = NRFX_RTC_INSTANCE(1);

// Wheter the slave periodically sends data to master.
bool sending_flag = false;

// Message counter
uint8_t msg_counter = 0;

/**
 *@brief Function for initializing logging.
 */
static void log_init(void)
{
    ret_code_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();
}

/** @brief: Function for handling the RTC1 interrupts.
 */
static void rtc_dummy_handler(nrfx_rtc_int_type_t int_type)
{
}

/** @brief Function for initialization and configuration of RTC driver instance.
 */
void rtc_init(void)
{
    uint32_t err_code;

    // Initialize RTC instance
    nrfx_rtc_config_t config = NRFX_RTC_DEFAULT_CONFIG;
    err_code = nrfx_rtc_init(&m_rtc, &config, rtc_dummy_handler);
    APP_ERROR_CHECK(err_code);

    // Power on RTC instance
    nrfx_rtc_enable(&m_rtc);
}

/**@brief Function for configuring the ANT channel.
 */
void antslave_ant_setup(void)
{
    uint32_t err_code;

    // Configure received message timestamp to use RTC1 (4 byte counter)
    ANT_TIME_STAMP_CONFIG stampConfig;
    stampConfig.ucTimeBase        = ANT_TIME_BASE_ALT1;
    stampConfig.bTimeStampEnabled = true;
    err_code                      = sd_ant_time_stamp_config_set(&stampConfig);
    APP_ERROR_CHECK(err_code);

    // ANT Channel init
    ant_channel_config_t channel_config =
    {
        .channel_number    = ANT_CHANNEL_NUMBER,
        .channel_type      = CHANNEL_TYPE_SLAVE, 
        .ext_assign        = EXT_TYPE,
        .rf_freq           = RF_FREQ,
        .transmission_type = CHAN_ID_TRANS_TYPE,
        .device_type       = CHAN_ID_DEV_TYPE,
        .device_number     = CHAN_ID_DEV_NUM,
        .channel_period    = CHAN_PERIOD,
        .network_number    = ANT_CHANNEL_DEFAULT_NETWORK,
    };

    err_code = ant_channel_init(&channel_config);
    APP_ERROR_CHECK(err_code);

    // Configure Adv. Burst
    uint8_t adv_burst_config[] = {ADV_BURST_MODE_ENABLE, ADV_BURST_MODES_SIZE_24_BYTES, 0, 0, 0, 0, 0, 0};
    err_code = sd_ant_adv_burst_config_set(adv_burst_config, sizeof(adv_burst_config));
    APP_ERROR_CHECK(err_code);

    // Set an infinite High priority search timeout.
    ant_search_config_t ant_search_config   = DEFAULT_ANT_SEARCH_CONFIG(ANT_CHANNEL_NUMBER);
    ant_search_config.high_priority_timeout = ANT_HIGH_PRIORITY_TIMEOUT_DISABLE;
    err_code = ant_search_init(&ant_search_config);
    APP_ERROR_CHECK(err_code);

    // Channel open
    err_code = sd_ant_channel_open(ANT_CHANNEL_NUMBER);
    APP_ERROR_CHECK(err_code);
}


/**
 * @brief Handle ANT events
 */
static void ant_event_handler(ant_evt_t * p_ant_evt, void * p_context)
{
    uint32_t err_code;
    switch (p_ant_evt->event)
    {
        case EVENT_TX:                      // Fall through
        case EVENT_TRANSFER_TX_COMPLETED:   // Fall through
        case EVENT_TRANSFER_TX_FAILED:
            if(sending_flag)
            {
                // Queue an ACK message for next channel period
                uint8_t buffer[8] = {0};
                err_code = sd_ant_acknowledge_message_tx(ANT_CHANNEL_NUMBER,
                                                         ANT_STANDARD_DATA_PAYLOAD_SIZE,
                                                         buffer);
                APP_ERROR_CHECK(err_code);

                // Queue an Adv. Burst message for next channel period
                // uint32_t buffer[32] = {0};
                // err_code = sd_ant_burst_handler_request(ANT_CHANNEL_NUMBER,
                //                                         32,
                //                                         buffer,
                //                                         BURST_SEGMENT_START | BURST_SEGMENT_END);
                // APP_ERROR_CHECK(err_code);
            }
        break;
        case EVENT_RX: 
            switch (p_ant_evt->message.ANT_MESSAGE_ucMesgID)
            {
                case MESG_ACKNOWLEDGED_DATA_ID:
                    if(!sending_flag)
                    {
                        // Raise the sending flag. Next msgs will be queue after the first one.
                        sending_flag = true;

                        // Queue the first ACK message for next channel period
                        uint8_t buffer[8] = {0};
                        err_code = sd_ant_acknowledge_message_tx(ANT_CHANNEL_NUMBER,
                                                                 ANT_STANDARD_DATA_PAYLOAD_SIZE,
                                                                 buffer);
                        APP_ERROR_CHECK(err_code);
                    }
                    
                break;
                case MESG_BROADCAST_DATA_ID:
                    // Print page number, user data and transmission time offset
                    if ((p_ant_evt->message.ANT_MESSAGE_aucPayload[0] == TIME_SYNC_PAGE) ||
                        (p_ant_evt->message.ANT_MESSAGE_aucPayload[0] == INVALIDATION_BYTE))
                    {
                        NRF_LOG_RAW_INFO("Page number: %lu\n", p_ant_evt->message.ANT_MESSAGE_aucPayload[0]); 
                        NRF_LOG_RAW_INFO("User data: %lu, %lu, %lu\n", p_ant_evt->message.ANT_MESSAGE_aucPayload[3],
                                                                       p_ant_evt->message.ANT_MESSAGE_aucPayload[4],
                                                                       p_ant_evt->message.ANT_MESSAGE_aucPayload[5]); 
                        NRF_LOG_RAW_INFO("Offset: %lu, %lu\n\n", p_ant_evt->message.ANT_MESSAGE_aucPayload[6],
                                                                   p_ant_evt->message.ANT_MESSAGE_aucPayload[7]); 
                    }
                break;
            }
        break;
        default:
        break;
    }
}

// Register the ant_evt_handler as an ANT observer.
NRF_SDH_ANT_OBSERVER(m_ant_observer, APP_ANT_OBSERVER_PRIO, ant_event_handler, NULL);

/**@brief Function for ANT stack initialization.
 */
void softdevice_setup(void)
{
    ret_code_t err_code = nrf_sdh_enable_request();
    APP_ERROR_CHECK(err_code);

    ASSERT(nrf_sdh_is_enabled());

    err_code = nrf_sdh_ant_enable();
    APP_ERROR_CHECK(err_code);
}

int main(void)
{
    // Log init
    log_init();

    // Setup SoftDevice and events handler      
    softdevice_setup();

    // RTC1 init.
    rtc_init();

    // Setup ANT and start channel
    antslave_ant_setup();

    // Enter main loop
    for (;;)
    {
        // Do nothing
    }
}


